#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2019-2020 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The ``astc_test_image_dl`` utility provides a means to programatically download
test images that are available online, avoiding the need to duplicate them in
the git repository.
"""


import os
import sys
import urllib.request

from PIL import Image


TEST_IMAGE_DIR = os.path.join("Test", "Images")


def download(testSet, index, srcUrl, dstPath):
    """
    Download a single image.

    Args:
        testSet (str): The test set name.
        index (int): The download index.
        srcUrl (str): The download URL.
        dstPath (str): The destination path.
    """
    dirName = os.path.dirname(dstPath)
    if not os.path.exists(dirName):
        os.makedirs(dirName)

    # Skip downloads if the file already exists
    if not os.path.exists(dstPath):
        print("%s image %u: Downloading" % (testSet, index))
        urllib.request.urlretrieve(srcUrl, dstPath)
    else:
        print("%s image %u: Skipping" % (testSet, index))


def make_landscape(imgPath):
    """
    Make an image on disk landscape aspect (edit in place)

    Args:
        imgPath: The pth of the image on disk.
    """
    img = Image.open(imgPath)
    if img.size[0] < img.size[1]:
        img = img.rotate(90, expand=True)
        img.save(imgPath)


def make_mixed_image(imgPathA, imgPathB, dstPath):
    """
    Make image consisting of RGB from A's RGB, and alpha from B's luminance.

    Args:
        imgPathA: The path of input A on disk.
        imgPathB: The path of input B on disk.
        dstPath: The path of the destination.
    """
    imgA = Image.open(imgPathA)
    imgB = Image.open(imgPathB).convert("L")

    imgA.putalpha(imgB)

    dirs = os.path.dirname(dstPath)
    if not os.path.exists(dirs):
        os.makedirs(dirs)

    imgA.save(dstPath)


def make_montage(imageDir, dstPath):
    """
    Make a single mosaic montage consisting of all of the Kodak images.

    Args:
        imgDir: The directory path of the Kodak images on disk.
        dstPth: The file path of the resulting montage.
    """
    cols = 6
    rows = 4

    width = 768
    height = 512

    images = os.listdir(imageDir)
    images.sort()

    montage = Image.new('RGB', (width * cols, height * rows))

    for i, src in enumerate(images):
        im = Image.open(os.path.join(imageDir, src))
        col = i % cols
        row = i // cols
        montage.paste(im, (width * col, height * row))

    dirs = os.path.dirname(dstPath)
    if not os.path.exists(dirs):
        os.makedirs(dirs)

    montage.save(dstPath)


def retrieve_kodak_set():
    """
    Download the public domain Kodak image set.

    To make test set mosaics easier to build we rotate images to make
    everything landscape.
    """
    testSet = "Kodak"

    # Download the original RGB images
    for i in range(1, 25):
        fle = "ldr-rgb-kodak%02u.png" % i
        dst = os.path.join(TEST_IMAGE_DIR, "Kodak", "LDR-RGB", fle)
        src = "http://r0k.us/graphics/kodak/kodak/kodim%02u.png" % i
        download(testSet, i, src, dst)

        # Canonicalize image aspect
        make_landscape(dst)

    # Make some correlated alpha RGBA images
    fle = "ldr-rgb-kodak%02u.png"  # Expand later
    pattern = os.path.join(TEST_IMAGE_DIR, "Kodak", "LDR-RGB", fle)

    for i in (22, 23):
        imgA = pattern % i
        fle = "ldr-rgba-kodak%02u+ca.png" % i
        dst = os.path.join(TEST_IMAGE_DIR, "KodakSim", "LDR-RGBA", fle)
        make_mixed_image(imgA, imgA, dst)

    # Make some non-correlated alpha RGBA images
    for i, j in ((22, 24), (23, 20)):
        imgA = pattern % i
        imgB = pattern % j
        fle = "ldr-rgba-kodak%02u+%02u+nca.png" % (i, j)
        dst = os.path.join(TEST_IMAGE_DIR, "KodakSim", "LDR-RGBA", fle)
        make_mixed_image(imgA, imgB, dst)

    # Make a large montage
    srcDir = os.path.join(TEST_IMAGE_DIR, "Kodak", "LDR-RGB")
    fle = "ldr-rgb-montage.png"
    dst = os.path.join(TEST_IMAGE_DIR, "KodakMnt", "LDR-RGB", fle)
    make_montage(srcDir, dst)


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    retrieve_kodak_set()
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2021 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The ``astc_trace_analysis`` utility provides a tool to analyze trace files.

WARNING: Trace files are an engineering tool, and not part of the standard
product, so traces and their associated tools are volatile and may change
significantly without notice.
"""

import argparse
from collections import defaultdict as ddict
import json
import numpy as np
import sys

QUANT_TABLE = {
	 0:   2,
	 1:   3,
	 2:   4,
	 3:   5,
	 4:   6,
	 5:   8,
	 6:  10,
	 7:  12,
	 8:  16,
	 9:  20,
	10:  24,
	11:  32
}

CHANNEL_TABLE = {
	 0: "R",
	 1: "G",
	 2: "B",
	 3: "A"
}

class Trace:

    def __init__(self, block_x, block_y, block_z):
        self.block_x = block_x
        self.block_y = block_y
        self.block_z = block_z
        self.blocks = []

    def add_block(self, block):
        self.blocks.append(block)

    def __getitem__(self, i):
       return self.blocks[i]

    def __delitem__(self, i):
       del self.blocks[i]

    def __len__(self):
       return len(self.blocks)

class Block:

    def __init__(self, pos_x, pos_y, pos_z, error_target):
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.pos_z = pos_z

        self.raw_min = None
        self.raw_max = None

        self.ldr_min = None
        self.ldr_max = None

        self.error_target = error_target
        self.passes = []
        self.qualityHit = None

    def add_minimums(self, r, g, b, a):
        self.raw_min = (r, g, b, a)

        def ldr(x):
            cmax = 65535.0
            return int((r / cmax) * 255.0)

        self.ldr_min = (ldr(r), ldr(g), ldr(b), ldr(a))

    def add_maximums(self, r, g, b, a):
        self.raw_max = (r, g, b, a)

        def ldr(x):
            cmax = 65535.0
            return int((r / cmax) * 255.0)

        self.ldr_max = (ldr(r), ldr(g), ldr(b), ldr(a))

    def add_pass(self, pas):
        self.passes.append(pas)

    def __getitem__(self, i):
       return self.passes[i]

    def __delitem__(self, i):
       del self.passes[i]

    def __len__(self):
       return len(self.passes)


class Pass:

    def __init__(self, partitions, partition, planes, target_hit, mode, component):
        self.partitions = partitions
        self.partition_index = 0 if partition is None else partition
        self.planes = planes
        self.plane2_component = component
        self.target_hit = target_hit
        self.search_mode = mode
        self.candidates = []

    def add_candidate(self, candidate):
        self.candidates.append(candidate)

    def __getitem__(self, i):
       return self.candidates[i]

    def __delitem__(self, i):
       del self.candidates[i]

    def __len__(self):
       return len(self.candidates)


class Candidate:

    def __init__(self, weight_x, weight_y, weight_z, weight_quant):
        self.weight_x = weight_x
        self.weight_y = weight_y
        self.weight_z = weight_z
        self.weight_quant = weight_quant
        self.refinement_errors = []

    def add_refinement(self, errorval):
        self.refinement_errors.append(errorval)


def get_attrib(data, name, multiple=False, hard_fail=True):
    results = []
    for attrib in data:
        if len(attrib) == 2 and attrib[0] == name:
            results.append(attrib[1])

    if not results:
        if hard_fail:
            print(json.dumps(data, indent=2))
            assert False, "Attribute %s not found" % name
        if multiple:
            return list()
        return None

    if not multiple:
        if len(results) > 1:
            print(json.dumps(data, indent=2))
            assert False, "Attribute %s found %u times" % (name, len(results))
        return results[0]

    return results


def rev_enumerate(seq):
    return zip(reversed(range(len(seq))), reversed(seq))

def foreach_block(data):

    for block in data:
        yield block

def foreach_pass(data):

    for block in data:
        for pas in block:
            yield (block, pas)

def foreach_candidate(data):

    for block in data:
        for pas in block:
            # Special case - None candidates for 0 partition
            if not len(pas):
                yield (block, pas, None)

            for candidate in pas:
                yield (block, pas, candidate)

def get_node(data, name, multiple=False, hard_fail=True):
    results = []
    for attrib in data:
        if len(attrib) == 3 and attrib[0] == "node" and attrib[1] == name:
            results.append(attrib[2])

    if not results:
        if hard_fail:
            print(json.dumps(data, indent=2))
            assert False, "Node %s not found" % name
        return None

    if not multiple:
        if len(results) > 1:
            print(json.dumps(data, indent=2))
            assert False, "Node %s found %u times" % (name, len(results))
        return results[0]

    return results


def find_best_pass_and_candidate(block):
    explicit_pass = None

    best_error = 1e30
    best_pass = None
    best_candidate = None

    for pas in block:
        # Special case for constant color blocks - no trial candidates
        if pas.target_hit and pas.partitions == 0:
            return (pas, None)

        for candidate in pas:
            errorval = candidate.refinement_errors[-1]
            if errorval <= best_error:
                best_error = errorval
                best_pass = pas
                best_candidate = candidate

    # Every other return type must have both best pass and best candidate
    assert (best_pass and best_candidate)
    return (best_pass, best_candidate)


def generate_database(data):
    # Skip header
    assert(data[0] == "node")
    assert(data[1] == "root")
    data = data[2]

    bx = get_attrib(data, "block_x")
    by = get_attrib(data, "block_y")
    bz = get_attrib(data, "block_z")
    dbStruct = Trace(bx, by, bz)

    for block in get_node(data, "block", True):
        px = get_attrib(block, "pos_x")
        py = get_attrib(block, "pos_y")
        pz = get_attrib(block, "pos_z")

        minr = get_attrib(block, "min_r")
        ming = get_attrib(block, "min_g")
        minb = get_attrib(block, "min_b")
        mina = get_attrib(block, "min_a")

        maxr = get_attrib(block, "max_r")
        maxg = get_attrib(block, "max_g")
        maxb = get_attrib(block, "max_b")
        maxa = get_attrib(block, "max_a")

        et = get_attrib(block, "tune_error_threshold")

        blockStruct = Block(px, py, pz, et)
        blockStruct.add_minimums(minr, ming, minb, mina)
        blockStruct.add_maximums(maxr, maxg, maxb, maxa)
        dbStruct.add_block(blockStruct)

        for pas in get_node(block, "pass", True):
            # Don't copy across passes we skipped due to heuristics
            skipped = get_attrib(pas, "skip", False, False)
            if skipped:
                continue

            prts = get_attrib(pas, "partition_count")
            prti = get_attrib(pas, "partition_index", False, False)
            plns = get_attrib(pas, "plane_count")
            chan = get_attrib(pas, "plane_component", False, plns > 2)
            mode = get_attrib(pas, "search_mode", False, False)
            ehit = get_attrib(pas, "exit", False, False) == "quality hit"

            passStruct = Pass(prts, prti, plns, ehit, mode, chan)
            blockStruct.add_pass(passStruct)

            # Constant color blocks don't have any candidates
            if prts == 0:
                continue

            for candidate in get_node(pas, "candidate", True):
                # Don't copy across candidates we couldn't encode
                failed = get_attrib(candidate, "failed", False, False)
                if failed:
                    continue

                wx = get_attrib(candidate, "weight_x")
                wy = get_attrib(candidate, "weight_y")
                wz = get_attrib(candidate, "weight_z")
                wq = QUANT_TABLE[get_attrib(candidate, "weight_quant")]
                epre = get_attrib(candidate, "error_prerealign", True, False)
                epst = get_attrib(candidate, "error_postrealign", True, False)

                candStruct = Candidate(wx, wy, wz, wq)
                passStruct.add_candidate(candStruct)
                for value in epre:
                    candStruct.add_refinement(value)
                for value in epst:
                    candStruct.add_refinement(value)

    return dbStruct


def filter_database(data):

    for block in data:
        best_pass, best_candidate = find_best_pass_and_candidate(block)

        for i, pas in rev_enumerate(block):
            if pas != best_pass:
                del block[i]
                continue

            if best_candidate is None:
                continue

            for j, candidate in rev_enumerate(pas):
                if candidate != best_candidate:
                    del pas[j]


def generate_pass_statistics(data):
    pass


def generate_feature_statistics(data):
    # -------------------------------------------------------------------------
    # Config
    print("Compressor Config")
    print("=================")

    if data.block_z > 1:
        dat = (data.block_x, data.block_y, data.block_z)
        print("  - Block size: %ux%ux%u" % dat)
    else:
        dat = (data.block_x, data.block_y)
        print("  - Block size: %ux%u" % dat)

    print("")

    # -------------------------------------------------------------------------
    # Block metrics
    result = ddict(int)

    RANGE_QUANT = 16

    for block in foreach_block(data):
        ranges = []
        for i in range(0, 4):
            ranges.append(block.ldr_max[i] - block.ldr_min[i])

        max_range = max(ranges)
        max_range = int(max_range / RANGE_QUANT) * RANGE_QUANT

        result[max_range] += 1

    print("Channel Range")
    print("=============")
    keys = sorted(result.keys())
    for key in keys:
        dat = (key, key + RANGE_QUANT - 1, result[key])
        print("  - %3u-%3u dynamic range = %6u blocks" % dat)

    print("")

    # -------------------------------------------------------------------------
    # Partition usage
    result_totals = ddict(int)
    results = ddict(lambda: ddict(int))

    for _, pas in foreach_pass(data):
        result_totals[pas.partitions] += 1
        results[pas.partitions][pas.partition_index] += 1

    print("Partition Count")
    print("===============")
    keys = sorted(result_totals.keys())
    for key in keys:
        dat = (key, result_totals[key], len(results[key]))
        print("  - %u partition(s) = %6u blocks / %4u indicies" % dat)

    print("")

    # -------------------------------------------------------------------------
    # Plane usage
    result_count = ddict(lambda: ddict(int))
    result_channel = ddict(lambda: ddict(int))

    for _, pas in foreach_pass(data):
        result_count[pas.partitions][pas.planes] += 1
        if (pas.planes > 1):
            result_channel[pas.partitions][pas.plane2_component] += 1

    print("Plane Usage")
    print("===========")
    keys = sorted(result_count.keys())
    for key in keys:
        keys2 = sorted(result_count[key])
        for key2 in keys2:
            val2 = result_count[key][key2]
            dat = (key, key2, val2)
            print("  - %u partition(s) %u plane(s) = %6u blocks" % dat)
            if key2 == 2:
                keys3 = sorted(result_channel[key])
                for key3 in keys3:
                    dat = (CHANNEL_TABLE[key3], result_channel[key][key3])
                    print("    - %s plane                 = %6u blocks" % dat)

    print("")

    # -------------------------------------------------------------------------
    # Decimation usage
    decim_count = ddict(lambda: ddict(int))
    quant_count = ddict(lambda: ddict(lambda: ddict(int)))


    MERGE_ROTATIONS = True

    for _, pas, can in foreach_candidate(data):
        # Skip constant color blocks
        if can is None:
            continue

        wx = can.weight_x
        wy = can.weight_y

        if MERGE_ROTATIONS and wx < wy:
            wx, wy = wy, wx

        decim_count[wx][wy] += 1
        quant_count[wx][wy][can.weight_quant] += 1

    print("Decimation Usage")
    print("================")

    if MERGE_ROTATIONS:
        print("  - Note: data merging grid rotations")

    x_keys = sorted(decim_count.keys())
    for x_key in x_keys:
        y_keys = sorted(decim_count[x_key])

        for y_key in y_keys:
            count = decim_count[x_key][y_key]
            dat = (x_key, y_key, count)
            print("  - %ux%u weights      = %6u blocks" % dat)

            q_keys = sorted(quant_count[x_key][y_key])
            for q_key in q_keys:
                count = quant_count[x_key][y_key][q_key]
                dat = (q_key, count)
                print("    - %2u quant range = %6u blocks" % dat)

    print("")

    # -------------------------------------------------------------------------
    # Refinement usage

    total_count = 0
    better_count = 0
    could_have_count = 0
    success_count = 0

    refinement_step = []

    for block, pas, candidate in foreach_candidate(data):
        # Ignore zero partition blocks - they don't use refinement
        if not candidate:
            continue

        target_error = block.error_target
        start_error = candidate.refinement_errors[0]
        end_error = candidate.refinement_errors[-1]

        rpf = float(start_error - end_error) / float(len(candidate.refinement_errors))
        rpf = abs(rpf)
        refinement_step.append(rpf / start_error)

        total_count += 1
        if end_error <= start_error:
            better_count += 1

        if end_error <= target_error:
            success_count += 1
        else:
            for refinement in candidate.refinement_errors:
                if refinement <= target_error:
                    could_have_count += 1
                    break


    print("Refinement Usage")
    print("================")
    print("  - %u refinements(s)" % total_count)
    print("  - %u refinements(s) improved" % better_count)
    print("  - %u refinements(s) worsened" % (total_count - better_count))
    print("  - %u refinements(s) could hit target, but didn't" % could_have_count)
    print("  - %u refinements(s) hit target" % success_count)
    print("  - %f mean step improvement" % np.mean(refinement_step))


def parse_command_line():
    """
    Parse the command line.

    Returns:
        Namespace: The parsed command line container.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("trace", type=argparse.FileType("r"),
                        help="The trace file to analyze")

    return parser.parse_args()


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    args = parse_command_line()

    data = json.load(args.trace)
    db = generate_database(data)
    filter_database(db)

    generate_feature_statistics(db)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2019-2022 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
This module contains code for loading image metadata from a file path on disk.

The directory path is structured:

    TestSetName/TestFormat/FileName

... and the file name is structured:

    colorProfile-colorFormat-name[-flags].extension
"""

from collections.abc import Iterable
import os
import re
import subprocess as sp

from PIL import Image as PILImage

import testlib.misc as misc


CONVERT_BINARY = ["convert"]


class ImageException(Exception):
    """
    Exception thrown for bad image specification.
    """


class TestImage():
    """
    Objects of this type contain metadata for a single test image on disk.

    Attributes:
        filePath: The path of the file on disk.
        outFilePath: The path of the output file on disk.
        testSet: The name of the test set.
        testFormat: The test format group.
        testFile: The test file name.
        colorProfile: The image compression color profile.
        colorFormat: The image color format.
        name: The image human name.
        is3D: True if the image is 3D, else False.
        isAlphaScaled: True if the image wants alpha scaling, else False.
        TEST_EXTS: Expected test image extensions.
        PROFILES: Tuple of valid color profile values.
        FORMATS: Tuple of valid color format values.
        FLAGS: Map of valid flags (key) and their meaning (value).
    """
    TEST_EXTS = (".jpg", ".png", ".tga", ".dds", ".hdr", ".ktx")

    PROFILES = ("ldr", "ldrs", "hdr")

    FORMATS = ("l", "la", "xy", "rgb", "rgba")

    FLAGS = {
        # Flags for image compression control
        "3": "3D image",
        "m": "Mask image",
        "a": "Alpha scaled image"
    }

    def __init__(self, filePath):
        """
        Create a new image definition, based on a structured file path.

        Args:
            filePath (str): The path of the image on disk.

        Raises:
            ImageException: The image couldn't be found or is unstructured.
        """
        self.filePath = os.path.abspath(filePath)
        if not os.path.exists(self.filePath):
            raise ImageException("Image doesn't exist (%s)" % filePath)

        # Decode the path
        scriptDir = os.path.dirname(__file__)
        rootInDir = os.path.join(scriptDir, "..", "Images")
        partialPath = os.path.relpath(self.filePath, rootInDir)
        parts = misc.path_splitall(partialPath)
        if len(parts) != 3:
            raise ImageException("Image path not path triplet (%s)" % parts)
        self.testSet = parts[0]
        self.testFormat = parts[1]
        self.testFile = parts[2]

        # Decode the file name
        self.decode_file_name(self.testFile)

        # Output file path (store base without extension)
        rootOutDir = os.path.join(scriptDir, "..", "..", "TestOutput")
        outFilePath = os.path.join(rootOutDir, partialPath)
        outFilePath = os.path.abspath(outFilePath)
        outFilePath = os.path.splitext(outFilePath)[0]
        self.outFilePath = outFilePath

    def decode_file_name(self, fileName):
        """
        Utility function to decode metadata from an encoded file name.

        Args:
            fileName (str): The file name to tokenize.

        Raises:
            ImageException: The image file path is badly structured.
        """
        # Strip off the extension
        rootName = os.path.splitext(fileName)[0]

        parts = rootName.split("-")

        # Decode the mandatory fields
        if len(parts) >= 3:
            self.colorProfile = parts[0]
            if self.colorProfile not in self.PROFILES:
                raise ImageException("Unknown color profile (%s)" % parts[0])

            self.colorFormat = parts[1]
            if self.colorFormat not in self.FORMATS:
                raise ImageException("Unknown color format (%s)" % parts[1])

            # Consistency check between directory and file names
            reencode = "%s-%s" % (self.colorProfile, self.colorFormat)
            compare = self.testFormat.lower()
            if reencode != compare:
                dat = (self.testFormat, reencode)
                raise ImageException("Mismatched test and image (%s:%s)" % dat)

            self.name = parts[2]

        # Set default values for the optional fields
        self.is3D = False
        self.isAlphaScaled = False

        # Decode the flags field if present
        if len(parts) >= 4:
            flags = parts[3]
            seenFlags = set()
            for flag in flags:
                if flag in seenFlags:
                    raise ImageException("Duplicate flag (%s)" % flag)
                if flag not in self.FLAGS:
                    raise ImageException("Unknown flag (%s)" % flag)
                seenFlags.add(flag)

            self.is3D = "3" in seenFlags
            self.isAlphaScaled = "a" in seenFlags

    def get_size(self):
        """
        Get the dimensions of this test image, if format is known.

        Known cases today where the format is not known:

        * 3D .dds files.
        * Any .ktx, .hdr, .exr, or .astc file.

        Returns:
            tuple(int, int): The dimensions of a 2D image, or ``None`` if PIL
            could not open the file.
        """
        try:
            img = PILImage.open(self.filePath)
        except IOError:
            # HDR files
            return None
        except NotImplementedError:
            # DDS files
            return None

        return (img.size[0], img.size[1])


class Image():
    """
    Wrapper around an image on the file system.
    """

    # TODO: We don't support KTX yet, as ImageMagick doesn't.
    SUPPORTED_LDR = ["bmp", "jpg", "png", "tga"]
    SUPPORTED_HDR = ["exr", "hdr"]

    @classmethod
    def is_format_supported(cls, fileFormat, profile=None):
        """
        Test if a given file format is supported by the library.

        Args:
            fileFormat (str): The file extension (excluding the ".").
            profile (str or None): The profile (ldr or hdr) of the image.

        Returns:
            bool: `True` if the image is supported, `False` otherwise.
        """
        assert profile in [None, "ldr", "hdr"]

        if profile == "ldr":
            return fileFormat in cls.SUPPORTED_LDR

        if profile == "hdr":
            return fileFormat in cls.SUPPORTED_HDR

        return fileFormat in cls.SUPPORTED_LDR or \
            fileFormat in cls.SUPPORTED_HDR

    def __init__(self, filePath):
        """
        Construct a new Image.

        Args:
            filePath (str): The path to the image on disk.
        """
        self.filePath = filePath
        self.proxyPath = None

    def get_colors(self, coords):
        """
        Get the image colors at the given coordinate.

        Args:
            coords (tuple or list): A single coordinate, or a list of
                coordinates to sample.

        Returns:
            tuple: A single sample color (if `coords` was a coordinate).
            list: A list of sample colors (if `coords` was a list).

            Colors are returned as float values between 0.0 and 1.0 for LDR,
            and float values which may exceed 1.0 for HDR.
        """
        colors = []

        # We accept both a list of positions and a single position;
        # canonicalize here so the main processing only handles lists
        isList = len(coords) != 0 and isinstance(coords[0], Iterable)

        if not isList:
            coords = [coords]

        for (x, y) in coords:
            command = list(CONVERT_BINARY)
            command += [self.filePath]

            # Ensure convert factors in format origin if needed
            command += ["-auto-orient"]

            command += [
                "-format", "%%[pixel:p{%u,%u}]" % (x, y),
                "info:"
            ]

            if os.name == 'nt':
                command.insert(0, "magick")

            result = sp.run(command, stdout=sp.PIPE, stderr=sp.PIPE,
                            check=True, universal_newlines=True)

            rawcolor = result.stdout.strip()

            # Decode ImageMagick's annoying named color outputs. Note that this
            # only handles "known" cases triggered by our test images, we don't
            # support the entire ImageMagick named color table.
            if rawcolor == "black":
                colors.append([0.0, 0.0, 0.0, 1.0])
            elif rawcolor == "white":
                colors.append([1.0, 1.0, 1.0, 1.0])
            elif rawcolor == "red":
                colors.append([1.0, 0.0, 0.0, 1.0])
            elif rawcolor == "blue":
                colors.append([0.0, 0.0, 1.0, 1.0])

            # Decode ImageMagick's format tuples
            elif rawcolor.startswith("srgba"):
                rawcolor = rawcolor[6:]
                rawcolor = rawcolor[:-1]
                channels = rawcolor.split(",")
                for i, channel in enumerate(channels):
                    if (i < 3) and channel.endswith("%"):
                        channels[i] = float(channel[:-1]) / 100.0
                    elif (i < 3) and not channel.endswith("%"):
                        channels[i] = float(channel) / 255.0
                    else:
                        channels[i] = float(channel)
                colors.append(channels)
            elif rawcolor.startswith("srgb"):
                rawcolor = rawcolor[5:]
                rawcolor = rawcolor[:-1]
                channels = rawcolor.split(",")
                for i, channel in enumerate(channels):
                    if (i < 3) and channel.endswith("%"):
                        channels[i] = float(channel[:-1]) / 100.0
                    if (i < 3) and not channel.endswith("%"):
                        channels[i] = float(channel) / 255.0
                channels.append(1.0)
                colors.append(channels)
            elif rawcolor.startswith("rgba"):
                rawcolor = rawcolor[5:]
                rawcolor = rawcolor[:-1]
                channels = rawcolor.split(",")
                for i, channel in enumerate(channels):
                    if (i < 3) and channel.endswith("%"):
                        channels[i] = float(channel[:-1]) / 100.0
                    elif (i < 3) and not channel.endswith("%"):
                        channels[i] = float(channel) / 255.0
                    else:
                        channels[i] = float(channel)
                colors.append(channels)
            elif rawcolor.startswith("rgb"):
                rawcolor = rawcolor[4:]
                rawcolor = rawcolor[:-1]
                channels = rawcolor.split(",")
                for i, channel in enumerate(channels):
                    if (i < 3) and channel.endswith("%"):
                        channels[i] = float(channel[:-1]) / 100.0
                    if (i < 3) and not channel.endswith("%"):
                        channels[i] = float(channel) / 255.0
                channels.append(1.0)
                colors.append(channels)
            else:
                print(x, y)
                print(rawcolor)
                assert False

        # ImageMagick decodes DDS files as BGRA not RGBA; manually correct
        if self.filePath.endswith("dds"):
            for color in colors:
                tmp = color[0]
                color[0] = color[2]
                color[2] = tmp

        # ImageMagick decodes EXR files with premult alpha; manually correct
        if self.filePath.endswith("exr"):
            for color in colors:
                color[0] /= color[3]
                color[1] /= color[3]
                color[2] /= color[3]

        # Undo list canonicalization if we were given a single scalar coord
        if not isList:
            return colors[0]

        return colors



# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2020-2022 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
A ResultSet stores a set of results about the performance of a TestSet. Each
set keeps result Records for each image and block size tested, that store the
PSNR and coding time.

ResultSets are often backed by a CSV file on disk, and a ResultSet can be
compared against a set of reference results created by an earlier test run.
"""


import csv
import enum
import numpy as np
import os


@enum.unique
class Result(enum.IntEnum):
    """
    An enumeration of test result status values.

    Attributes:
        NOTRUN: The test has not been run.
        PASS: The test passed.
        WARN: The test image quality was below the pass threshold but above
            the fail threshold.
        FAIL: The test image quality was below the fail threshold.
    """
    NOTRUN = 0
    PASS = 1
    WARN = 2
    FAIL = 3


class ResultSummary():
    """
    An result summary data container, storing number of results of each type.

    Attributes:
        notruns: The number of tests that did not run.
        passes: The number of tests that passed.
        warnings: The number of tests that produced a warning.
        fails: The number of tests that failed.
        tTimes: Total time speedup vs reference (<1 is slower, >1 is faster).
        cTimes: Coding time speedup vs reference (<1 is slower, >1 is faster).
        psnrs: Coding time quality vs reference (<0 is worse, >0 is better).
    """

    def __init__(self):
        """
        Create a new result summary.
        """
        # Pass fail metrics
        self.notruns = 0
        self.passes = 0
        self.warnings = 0
        self.fails = 0

        # Relative results
        self.tTimesRel = []
        self.cTimesRel = []
        self.psnrRel = []

        # Absolute results
        self.cTime = []
        self.psnr = []

    def add_record(self, record):
        """
        Add a record to this summary.

        Args:
            record (Record): The Record to add.
        """
        if record.status == Result.PASS:
            self.passes += 1
        elif record.status == Result.WARN:
            self.warnings += 1
        elif record.status == Result.FAIL:
            self.fails += 1
        else:
            self.notruns += 1

        if record.tTimeRel is not None:
            self.tTimesRel.append(record.tTimeRel)
            self.cTimesRel.append(record.cTimeRel)
            self.psnrRel.append(record.psnrRel)

            self.cTime.append(record.cTime)
            self.psnr.append(record.psnr)

    def get_worst_result(self):
        """
        Get the worst result in this set.

        Returns:
            Result: The worst test result.
        """
        if self.fails:
            return Result.FAIL

        if self.warnings:
            return Result.WARN

        if self.passes:
            return Result.PASS

        return Result.NOTRUN

    def __str__(self):
        # Overall pass/fail results
        overall = self.get_worst_result().name
        dat = (overall, self.passes, self.warnings, self.fails)
        result = ["\nSet Status: %s (Pass: %u | Warn: %u | Fail: %u)" % dat]

        if (self.tTimesRel):
            # Performance summaries
            dat = (np.mean(self.tTimesRel), np.std(self.tTimesRel))
            result.append("\nTotal speed:   Mean:  %+0.3f x   Std: %0.2f x" % dat)

            dat = (np.mean(self.cTimesRel), np.std(self.cTimesRel))
            result.append("Coding speed:  Mean:  %+0.3f x   Std: %0.2f x" % dat)

            dat = (np.mean(self.psnrRel), np.std(self.psnrRel))
            result.append("Quality diff:  Mean:  %+0.3f dB  Std: %0.2f dB" % dat)

            dat = (np.mean(self.cTime), np.std(self.cTime))
            result.append("Coding time:   Mean:  %+0.3f s   Std: %0.2f s" % dat)

            dat = (np.mean(self.psnr), np.std(self.psnr))
            result.append("Quality:       Mean: %+0.3f dB  Std: %0.2f dB" % dat)

        return "\n".join(result)


class Record():
    """
    A single result record, sotring results for a singel image and block size.

    Attributes:
        blkSz: The block size.
        name: The test image name.
        psnr: The image quality (PSNR dB)
        tTime: The total compression time.
        cTime: The coding compression time.
        cRate: The coding compression rate.
        status: The test Result.
    """

    def __init__(self, blkSz, name, psnr, tTime, cTime, cRate):
        """
        Create a result record, initially in the NOTRUN status.

        Args:
            blkSz (str): The block size.
            name (str): The test image name.
            psnr (float): The image quality PSNR, in dB.
            tTime (float): The total compression time, in seconds.
            cTime (float): The coding compression time, in seconds.
            cRate (float): The coding compression rate, in MPix/s.
            tTimeRel (float): The relative total time speedup vs reference.
            cTimeRel (float): The relative coding time speedup vs reference.
            cRateRel (float): The relative rate speedup vs reference.
            psnrRel (float): The relative PSNR dB vs reference.
        """
        self.blkSz = blkSz
        self.name = name
        self.psnr = psnr
        self.tTime = tTime
        self.cTime = cTime
        self.cRate = cRate
        self.status = Result.NOTRUN

        self.tTimeRel = None
        self.cTimeRel = None
        self.cRateRel = None
        self.psnrRel = None

    def set_status(self, result):
        """
        Set the result status.

        Args:
            result (Result): The test result.
        """
        self.status = result

    def __str__(self):
        return "'%s' / '%s'" % (self.blkSz, self.name)


class ResultSet():
    """
    A set of results for a TestSet, across one or more block sizes.

    Attributes:
        testSet: The test set these results are linked to.
        records: The list of test results.
    """

    def __init__(self, testSet):
        """
        Create a new empty ResultSet.

        Args:
            testSet (TestSet): The test set these results are linked to.
        """
        self.testSet = testSet
        self.records = []

    def add_record(self, record):
        """
        Add a new test record to this result set.

        Args:
            record (Record): The test record to add.
        """
        self.records.append(record)

    def get_record(self, testSet, blkSz, name):
        """
        Get a record matching the arguments.

        Args:
            testSet (TestSet): The test set to get results from.
            blkSz (str): The block size.
            name (str): The test name.

        Returns:
            Record: The test result, if present.

        Raises:
            KeyError: No match could be found.
        """
        if testSet != self.testSet:
            raise KeyError()

        for record in self.records:
            if record.blkSz == blkSz and record.name == name:
                return record

        raise KeyError()

    def get_matching_record(self, other):
        """
        Get a record matching the config of another record.

        Args:
            other (Record): The pattern result record to match.

        Returns:
            Record: The result, if present.

        Raises:
            KeyError: No match could be found.
        """
        for record in self.records:
            if record.blkSz == other.blkSz and record.name == other.name:
                return record

        raise KeyError()

    def get_results_summary(self):
        """
        Get a results summary of all the records in this result set.

        Returns:
            ResultSummary: The result summary.
        """
        summary = ResultSummary()
        for record in self.records:
            summary.add_record(record)

        return summary

    def save_to_file(self, filePath):
        """
        Save this result set to a CSV file.

        Args:
            filePath (str): The output file path.
        """
        dirName = os.path.dirname(filePath)
        if not os.path.exists(dirName):
            os.makedirs(dirName)

        with open(filePath, "w", newline="") as csvfile:
            writer = csv.writer(csvfile)
            self._save_header(writer)
            for record in self.records:
                self._save_record(writer, record)

    @staticmethod
    def _save_header(writer):
        """
        Write the header to the CSV file.

        Args:
            writer (csv.writer): The CSV writer.
        """
        row = ["Image Set", "Block Size", "Name",
               "PSNR", "Total Time", "Coding Time", "Coding Rate"]
        writer.writerow(row)

    def _save_record(self, writer, record):
        """
        Write a record to the CSV file.

        Args:
            writer (csv.writer): The CSV writer.
            record (Record): The record to write.
        """
        row = [self.testSet,
               record.blkSz,
               record.name,
               "%0.4f" % record.psnr,
               "%0.4f" % record.tTime,
               "%0.4f" % record.cTime,
               "%0.4f" % record.cRate]
        writer.writerow(row)

    def load_from_file(self, filePath):
        """
        Load a reference result set from a CSV file on disk.

        Args:
            filePath (str): The input file path.
        """
        with open(filePath, "r") as csvfile:
            reader = csv.reader(csvfile)
            # Skip the header
            next(reader)
            for row in reader:
                assert row[0] == self.testSet
                record = Record(row[1], row[2],
                                float(row[3]), float(row[4]),
                                float(row[5]), float(row[6]))
                self.add_record(record)



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2022 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
This script is a simple test runner for sweeps on multiple compressors.
"""

import os
import subprocess as sp
import re
import sys

LOG_COMMANDS = False
LOG_PATTERN = re.compile(r"\s*Coding rate:\s*(.*)\s*MT/s")

ISPC_BIN = "./Binaries/ISPC/ispc_astc.exe"
ISPC_QUALITY = ["rgba", "rgb"]

ASTC_BIN = "./bin/astcenc-avx2"
ASTC_QUALITY = ["0", "8", "10", "20", "30", "40", "50", "60"]

TEST_BLOCK_SIZES = ["4x4", "6x6", "8x8"]

TEST_IMAGE = "./Test/Images/Kodak/LDR-RGB/ldr-rgb-kodak%02u.png"
TEST_RANGE = 24
TEST_REPEATS = 5

OUT_CIMAGE = "out.astc"
OUT_DIMAGE = "out.png"


def run(command):
    if LOG_COMMANDS:
        print(" ".join(command))

    return sp.run(command, capture_output=True, universal_newlines=True)


def run_astcenc(in_image, out_image, block_size, quality):
    args = [ASTC_BIN, "-tl", in_image, out_image, block_size, quality, "-j", "1"]
    result = run(args)
    return float(LOG_PATTERN.search(result.stdout).group(1))


def run_ispc(in_image, out_image, block_size, quality):
    args = [ISPC_BIN, in_image, out_image, block_size, quality]
    result = run(args)
    return float(LOG_PATTERN.search(result.stdout).group(1))


def decompress(in_image, out_image):
    args = [ASTC_BIN, "-dl", in_image, out_image]
    result = run(args)
    os.remove(in_image)


def compare(in_image, out_image):
    args = ["compare", "-metric", "PSNR", in_image, out_image, "diff.png"]
    result = run(args)
    os.remove("diff.png")
    os.remove(out_image)
    return float(result.stderr)


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    # ISPC Tests
    for block_size in TEST_BLOCK_SIZES:
        for quality in ISPC_QUALITY:
            print(f"ISPC {quality} {block_size}")
            print(f"ISPC {quality} {block_size}", file=sys.stderr)
            for index in range(1, TEST_RANGE + 1):
                result_rate = 0.0
                for repeat in range(0, TEST_REPEATS):
                    image = TEST_IMAGE % index
                    result_rate += run_ispc(image, OUT_CIMAGE, block_size, quality)
                    decompress(OUT_CIMAGE, OUT_DIMAGE)
                    result_error = compare(image, OUT_DIMAGE)
                result_rate /= TEST_REPEATS

                print("%s,Kodak%02u,%0.4f,%0.4f" % (block_size, index, result_rate, result_error))

    # ASTCENC Tests
    for block_size in TEST_BLOCK_SIZES:
        for quality in ASTC_QUALITY:
            print(f"ASTC {quality} {block_size}")
            print(f"ASTC {quality} {block_size}", file=sys.stderr)
            for index in range(1, TEST_RANGE + 1):
                result_rate = 0.0
                for repeat in range(0, TEST_REPEATS):
                    image = TEST_IMAGE % index
                    result_rate += run_astcenc(image, OUT_DIMAGE, block_size, quality)
                    result_error = compare(image, OUT_DIMAGE)
                result_rate /= TEST_REPEATS

                print("%s,Kodak%02u,%0.4f,%0.4f" % (block_size, index, result_rate, result_error))

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except sp.CalledProcessError as ex:
        print(ex.stdout)
        print(ex.stderr)



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2021-2022 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
A benchmarking sweep helper, which can generate a performance-vs-quality sweep
for a single input images. Like other test functionality, this uses structured
image directory layouts for determining image settings to pass to the codec.
"""

import argparse
import os
import platform
import sys

import testlib.encoder as te
import testlib.image as ti


def parse_command_line():
    """
    Parse the command line.

    Returns:
        Namespace: The parsed command line container.
    """
    parser = argparse.ArgumentParser()

    # All reference encoders
    parser.add_argument("--step", dest="step", default="10", type=int, help="step size")
    parser.add_argument("--repeats", dest="repeats", type=int, default=1, help="repeats")

    parser.add_argument(dest="image", default=None,
                        help="select the test image to run")

    args = parser.parse_args()
    return args


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    # Parse command lines
    args = parse_command_line()

    blockSizes = ["4x4", "5x5", "6x6", "8x8", "10x10"]
    repeats = max(args.repeats, 1)
    step = max(args.step, 1)

    image = ti.TestImage(args.image)
    codec = te.Encoder2x("avx2")

    print("Block Size, Quality, PSNR (dB), Coding Time (s), Coding Rate (MT/s)")

    for blockSize in blockSizes:
        for quality in range(0, 101, args.step):
            localRepeats = repeats
            if quality < 20:
                localRepeats = localRepeats * 2
            if quality < 40:
                localRepeats = localRepeats * 2

            results = codec.run_test(image, blockSize, f"{quality}", localRepeats, False)
            psnr = results[0]
            codingTime = results[2]
            mts = results[3]

            print(f"{blockSize}, {quality}, {psnr}, {codingTime}, {mts}")

    return 0


if __name__ == "__main__":
    sys.exit(main())



# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2020 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
An ASTC TestSet is comprised of a set of TestImages. Images are stored in a
structured directory layout. This structure encodes important metadata about
each image - such as color profile and data encoding - in the directory and
file names used.

TestSets are built by using reflection on a root directory to automatically
find all of the test images that comprise the set.
"""

import os

from testlib.image import TestImage


class TSetException(Exception):
    """
    Exception thrown for bad test set specification.
    """


class TestSet():
    """
    Generate a list of images that are test candidates.

    This reflection is built automatically based on a directory of images on
    disk, provided that the images follow a standard structure.

    Attributes:
        name: The name of the test set.
        tests: The list of TestImages forming the set.
    """

    def __init__(self, name, rootDir, profiles, formats, imageFilter=None):
        """
        Create a new TestSet through reflection.

        Args:
            name (str): The name of the test set.
            rootDir (str): The root directory of the test set.
            profiles (list(str)): The ASTC profiles to allow.
            formats (list(str)): The image formats to allow.
            imageFilter (str): The name of the image to include (for bug repo).

        Raises:
            TSetException: The specified TestSet could not be loaded.
        """
        self.name = name

        if not os.path.exists(rootDir) and not os.path.isdir(rootDir):
            raise TSetException("Bad test set root directory (%s)" % rootDir)

        self.tests = []

        for (dirPath, dirNames, fileNames) in os.walk(rootDir):
            for fileName in fileNames:
                # Only select image files
                fileExt = os.path.splitext(fileName)[1]
                if fileExt not in TestImage.TEST_EXTS:
                    continue

                # Create the TestImage for each file on disk
                filePath = os.path.join(dirPath, fileName)
                image = TestImage(filePath)

                # Filter out the ones we don't want to allow
                if image.colorProfile not in profiles:
                    continue

                if image.colorFormat not in formats:
                    continue

                if imageFilter and image.testFile != imageFilter:
                    continue

                self.tests.append((filePath, image))

        # Sort the TestImages so they are in a stable order
        self.tests.sort()
        self.tests = [x[1] for x in self.tests]



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2020 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The ``astc_image_info`` utility provides basic image query capabilities. It is
a modal command line utility, exposing multiple available operators.

* ``info``: Query structural information about the image, such as image
      dimensions, number of color channels, and the min/max of each channel.
* ``color``: Query the stored color value at a specific pixel coordinate, and
      print the result in a variety of different formats.

Both modes allow multiple images to be specified on the command line.
"""

import argparse
import sys

from PIL import Image


def main_color(args):
    """
    Main function for the "color" mode.

    This mode prints the color at a specific pixel coordinate in each image.
    The color value is printed in a variety of color formats (decimal, HTML
    string, float).

    Args:
        args (Namespace): The parsed command line arguments.

    Returns:
        int: The process return code.
    """
    retCode = 0

    for i, image in enumerate(args.images):
        if i != 0:
            print("")

        img = Image.open(image.name)
        x = args.location[0]
        y = args.location[1]

        print(image.name)
        print("=" * len(image.name))

        if (x >= img.size[0]) or (y >= img.size[1]):
            print("- ERROR: location out-of-bounds [%ux%u]" % img.size)
            retCode = 1
        else:
            color = img.getpixel((x, y))

            # Print byte values
            print("+ Byte: %s" % str(color))

            # Print hex values
            fmtString = "+ Hex: #" + ("%02X" * len(color))
            print(fmtString % color)

            # Print float values
            parts = ["%g"] * len(color)
            parts = ", ".join(parts)
            fmtString = "+ Float: (" + parts + ")"
            print(fmtString % tuple(float(x)/255.0 for x in color))

    return retCode


def main_info(args):
    """
    Main function for the "info" mode.

    This mode prints the basic metadata of an image:

        - the overall image size.
        - the number of color channels.
        - the min/max value in each color channel.

    Args:
        args (Namespace): The parsed command line arguments.

    Returns:
        int: The process return code.
    """
    for i, image in enumerate(args.images):
        if i != 0:
            print("")

        img = Image.open(image.name)
        minmax = img.getextrema()

        print(image.name)
        print("=" * len(image.name))
        print("+ Size: %ux%u" % (img.size[0], img.size[1]))
        print("+ Channels: %s" % ("".join(img.getbands())))
        for j, channel in enumerate(img.getbands()):
            print("  + %s: %u - %u" % (channel, *minmax[j]))

    return 0


def parse_loc(value):
    """
    Command line argument parser for position arguments.

    Args:
        value (str): The command line argument string to parse. Must be of the
            form <int>x<int>", where both integers must be zero or positive.

    Returns:
        list(int, int): The parsed location.

    Raises:
        ArgumentTypeError: The value is not a valid location.
    """
    error = argparse.ArgumentTypeError("%s is an invalid location" % value)
    svalue = value.split("x")

    if len(svalue) != 2:
        raise error

    try:
        ivalue = [int(x) for x in svalue if int(x) >= 0]
    except ValueError:
        raise error

    if len(ivalue) != len(svalue):
        raise error

    return ivalue


def parse_command_line():
    """
    Parse the command line.

    Returns:
        Namespace: The parsed command line container.
    """
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        title="Operations")

    # Create the parser for the "pipette" command
    parserA = subparsers.add_parser(
        "color",
        help="Print color at given coordinate")

    parserA.set_defaults(func=main_color)

    parserA.add_argument(
        "location", metavar="loc", type=parse_loc,
        help="The location spec XxY")

    parserA.add_argument(
        "images", metavar="image", nargs="+", type=argparse.FileType("r"),
        help="The images to query")

    # Create the parser for the "size" command
    parserB = subparsers.add_parser(
        "info",
        help="Print image metadata info")

    parserB.set_defaults(func=main_info)

    parserB.add_argument(
        "images", metavar="image", nargs="+", type=argparse.FileType("r"),
        help="The images to query")

    # Cope with the user failing to specify any sub-command. Note on Python 3.8
    # we could use required=True on the add_subparsers call, but we cannot do
    # this on 3.6 which is our current min-spec.
    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.print_help()
        return None

    return args


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    args = parse_command_line()
    if args:
        return args.func(args)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2019-2023 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
These classes provide an abstraction around the astcenc command line tool,
allowing the rest of the image test suite to ignore changes in the command line
interface.
"""

import os
import re
import subprocess as sp
import sys


class EncoderBase():
    """
    This class is a Python wrapper for the `astcenc` binary, providing an
    abstract means to set command line options and parse key results.

    This is an abstract base class providing some generic helper functionality
    used by concrete instantiations of subclasses.

    Attributes:
        binary: The encoder binary path.
        variant: The encoder SIMD variant being tested.
        name: The encoder name to use in reports.
        VERSION: The encoder version or branch.
        SWITCHES: Dict of switch replacements for different color formats.
        OUTPUTS: Dict of output file extensions for different color formats.
    """

    VERSION = None
    SWITCHES = None
    OUTPUTS = None

    def __init__(self, name, variant, binary):
        """
        Create a new encoder instance.

        Args:
            name (str): The name of the encoder.
            variant (str): The SIMD variant of the encoder.
            binary (str): The path to the binary on the file system.
        """
        self.name = name
        self.variant = variant
        self.binary = binary

    def build_cli(self, image, blockSize="6x6", preset="-thorough",
                  keepOutput=True, threads=None):
        """
        Build the command line needed for the given test.

        Args:
            image (TestImage): The test image to compress.
            blockSize (str): The block size to use.
            preset (str): The quality-performance preset to use.
            keepOutput (bool): Should the test preserve output images? This is
                only a hint and discarding output may be ignored if the encoder
                version used can't do it natively.
            threads (int or None): The thread count to use.

        Returns:
            list(str): A list of command line arguments.
        """
        # pylint: disable=unused-argument,no-self-use,redundant-returns-doc
        assert False, "Missing subclass implementation"

    def execute(self, command):
        """
        Run a subprocess with the specified command.

        Args:
            command (list(str)): The list of command line arguments.

        Returns:
            list(str): The output log (stdout) split into lines.
        """
        # pylint: disable=no-self-use
        try:
            result = sp.run(command, stdout=sp.PIPE, stderr=sp.PIPE,
                            check=True, universal_newlines=True)
        except (OSError, sp.CalledProcessError):
            print("ERROR: Test run failed")
            print("  + %s" % " ".join(command))
            qcommand = ["\"%s\"" % x for x in command]
            print("  + %s" % ", ".join(qcommand))
            sys.exit(1)

        return result.stdout.splitlines()

    def parse_output(self, image, output):
        """
        Parse the log output for PSNR and performance metrics.

        Args:
            image (TestImage): The test image to compress.
            output (list(str)): The output log from the compression process.

        Returns:
            tuple(float, float, float): PSNR in dB, TotalTime in seconds, and
            CodingTime in seconds.
        """
        # Regex pattern for image quality
        patternPSNR = re.compile(self.get_psnr_pattern(image))
        patternTTime = re.compile(self.get_total_time_pattern())
        patternCTime = re.compile(self.get_coding_time_pattern())
        patternCRate = re.compile(self.get_coding_rate_pattern())

        # Extract results from the log
        runPSNR = None
        runTTime = None
        runCTime = None
        runCRate = None

        for line in output:
            match = patternPSNR.match(line)
            if match:
                runPSNR = float(match.group(1))

            match = patternTTime.match(line)
            if match:
                runTTime = float(match.group(1))

            match = patternCTime.match(line)
            if match:
                runCTime = float(match.group(1))

            match = patternCRate.match(line)
            if match:
                runCRate = float(match.group(1))

        stdout = "\n".join(output)
        assert runPSNR is not None, "No coding PSNR found %s" % stdout
        assert runTTime is not None, "No total time found %s" % stdout
        assert runCTime is not None, "No coding time found %s" % stdout
        assert runCRate is not None, "No coding rate found %s" % stdout
        return (runPSNR, runTTime, runCTime, runCRate)

    def get_psnr_pattern(self, image):
        """
        Get the regex pattern to match the image quality metric.

        Note, while this function is called PSNR for some images we may choose
        to match another metric (e.g. mPSNR for HDR images).

        Args:
            image (TestImage): The test image we are compressing.

        Returns:
            str: The string for a regex pattern.
        """
        # pylint: disable=unused-argument,no-self-use,redundant-returns-doc
        assert False, "Missing subclass implementation"

    def get_total_time_pattern(self):
        """
        Get the regex pattern to match the total compression time.

        Returns:
            str: The string for a regex pattern.
        """
        # pylint: disable=unused-argument,no-self-use,redundant-returns-doc
        assert False, "Missing subclass implementation"

    def get_coding_time_pattern(self):
        """
        Get the regex pattern to match the coding compression time.

        Returns:
            str: The string for a regex pattern.
        """
        # pylint: disable=unused-argument,no-self-use,redundant-returns-doc
        assert False, "Missing subclass implementation"

    def run_test(self, image, blockSize, preset, testRuns, keepOutput=True,
                 threads=None):
        """
        Run the test N times.

        Args:
            image (TestImage): The test image to compress.
            blockSize (str): The block size to use.
            preset (str): The quality-performance preset to use.
            testRuns (int): The number of test runs.
            keepOutput (bool): Should the test preserve output images? This is
                only a hint and discarding output may be ignored if the encoder
                version used can't do it natively.
            threads (int or None): The thread count to use.

        Returns:
            tuple(float, float, float, float): Returns the best results from
            the N test runs, as PSNR (dB), total time (seconds), coding time
            (seconds), and coding rate (M pixels/s).
        """
        # pylint: disable=assignment-from-no-return
        command = self.build_cli(image, blockSize, preset, keepOutput, threads)

        # Execute test runs
        bestPSNR = 0
        bestTTime = sys.float_info.max
        bestCTime = sys.float_info.max
        bestCRate = 0

        for _ in range(0, testRuns):
            output = self.execute(command)
            result = self.parse_output(image, output)

            # Keep the best results (highest PSNR, lowest times, highest rate)
            bestPSNR = max(bestPSNR, result[0])
            bestTTime = min(bestTTime, result[1])
            bestCTime = min(bestCTime, result[2])
            bestCRate = max(bestCRate, result[3])

        return (bestPSNR, bestTTime, bestCTime, bestCRate)


class Encoder2x(EncoderBase):
    """
    This class wraps the latest `astcenc` 2.x series binaries from main branch.
    """
    VERSION = "main"

    SWITCHES = {
        "ldr": "-tl",
        "ldrs": "-ts",
        "hdr": "-th",
        "hdra": "-tH"
    }

    OUTPUTS = {
        "ldr": ".png",
        "ldrs": ".png",
        "hdr": ".exr",
        "hdra": ".exr"
    }

    def __init__(self, variant, binary=None):
        name = "astcenc-%s-%s" % (variant, self.VERSION)

        if binary is None:
            if variant != "universal":
                binary = f"./bin/astcenc-{variant}"
            else:
                binary = "./bin/astcenc"

            if os.name == 'nt':
                binary = f"{binary}.exe"

        super().__init__(name, variant, binary)

    def build_cli(self, image, blockSize="6x6", preset="-thorough",
                  keepOutput=True, threads=None):
        opmode = self.SWITCHES[image.colorProfile]
        srcPath = image.filePath

        if keepOutput:
            dstPath = image.outFilePath + self.OUTPUTS[image.colorProfile]
            dstDir = os.path.dirname(dstPath)
            dstFile = os.path.basename(dstPath)
            dstPath = os.path.join(dstDir, self.name, preset[1:], blockSize, dstFile)

            dstDir = os.path.dirname(dstPath)
            os.makedirs(dstDir, exist_ok=True)
        elif sys.platform == "win32":
            dstPath = "nul"
        else:
            dstPath = "/dev/null"

        command = [
            self.binary, opmode, srcPath, dstPath,
            blockSize, preset, "-silent"
        ]

        if image.colorFormat == "xy":
            command.append("-normal")

        if image.isAlphaScaled:
            command.append("-a")
            command.append("1")

        if threads is not None:
            command.append("-j")
            command.append("%u" % threads)

        return command

    def get_psnr_pattern(self, image):
        if image.colorProfile != "hdr":
            if image.colorFormat != "rgba":
                patternPSNR = r"\s*PSNR \(LDR-RGB\):\s*([0-9.]*) dB"
            else:
                patternPSNR = r"\s*PSNR \(LDR-RGBA\):\s*([0-9.]*) dB"
        else:
            patternPSNR = r"\s*mPSNR \(RGB\)(?: \[.*?\] )?:\s*([0-9.]*) dB.*"
        return patternPSNR

    def get_total_time_pattern(self):
        return r"\s*Total time:\s*([0-9.]*) s"

    def get_coding_time_pattern(self):
        return r"\s*Coding time:\s*([0-9.]*) s"

    def get_coding_rate_pattern(self):
        return r"\s*Coding rate:\s*([0-9.]*) MT/s"


class Encoder2xRel(Encoder2x):
    """
    This class wraps a released 2.x series binary.
    """
    def __init__(self, version, variant):

        self.VERSION = version

        if variant != "universal":
            binary = f"./Binaries/{version}/astcenc-{variant}"
        else:
            binary = f"./Binaries/{version}/astcenc"

        if os.name == 'nt':
            binary = f"{binary}.exe"

        super().__init__(variant, binary)


class Encoder1_7(EncoderBase):
    """
    This class wraps the 1.7 series binaries.
    """
    VERSION = "1.7"

    SWITCHES = {
        "ldr": "-tl",
        "ldrs": "-ts",
        "hdr": "-t"
    }

    OUTPUTS = {
        "ldr": ".tga",
        "ldrs": ".tga",
        "hdr": ".htga"
    }

    def __init__(self):
        name = "astcenc-%s" % self.VERSION
        if os.name == 'nt':
            binary = "./Binaries/1.7/astcenc.exe"
        else:
            binary = "./Binaries/1.7/astcenc"

        super().__init__(name, None, binary)

    def build_cli(self, image, blockSize="6x6", preset="-thorough",
                  keepOutput=True, threads=None):

        if preset == "-fastest":
            preset = "-fast"

        opmode = self.SWITCHES[image.colorProfile]
        srcPath = image.filePath

        dstPath = image.outFilePath + self.OUTPUTS[image.colorProfile]
        dstDir = os.path.dirname(dstPath)
        dstFile = os.path.basename(dstPath)
        dstPath = os.path.join(dstDir, self.name, preset[1:], blockSize, dstFile)

        dstDir = os.path.dirname(dstPath)
        os.makedirs(dstDir, exist_ok=True)

        command = [
            self.binary, opmode, srcPath, dstPath,
            blockSize, preset, "-silentmode", "-time", "-showpsnr"
        ]

        if image.colorFormat == "xy":
            command.append("-normal_psnr")

        if image.colorProfile == "hdr":
            command.append("-hdr")

        if image.isAlphaScaled:
            command.append("-alphablend")

        if threads is not None:
            command.append("-j")
            command.append("%u" % threads)

        return command

    def get_psnr_pattern(self, image):
        if image.colorProfile != "hdr":
            if image.colorFormat != "rgba":
                patternPSNR = r"PSNR \(LDR-RGB\):\s*([0-9.]*) dB"
            else:
                patternPSNR = r"PSNR \(LDR-RGBA\):\s*([0-9.]*) dB"
        else:
            patternPSNR = r"mPSNR \(RGB\)(?: \[.*?\] )?:\s*([0-9.]*) dB.*"
        return patternPSNR

    def get_total_time_pattern(self):
        # Pattern match on a new pattern for a 2.1 compatible variant
        # return r"Elapsed time:\s*([0-9.]*) seconds.*"
        return r"\s*Total time:\s*([0-9.]*) s"

    def get_coding_time_pattern(self):
        # Pattern match on a new pattern for a 2.1 compatible variant
        # return r".* coding time: \s*([0-9.]*) seconds"
        return r"\s*Coding time:\s*([0-9.]*) s"

    def get_coding_rate_pattern(self):
        # Pattern match on a new pattern for a 2.1 compatible variant
        return r"\s*Coding rate:\s*([0-9.]*) MT/s"



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2021-2022 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The ``astc_dump_binary`` utility provides a wrapper around the ``objdump``
utility to extract disassembly of specific functions. Currently only matches
the root name, for sake of command line sanity, so all overloads get dumped.

Using __attribute__((noinline)) can be useful during profiling to stop any
functions of interest getting inlined once they get too small ...
"""


import argparse
import re
import shutil
import subprocess as sp
import sys


def run_objdump(binary, symbol):
    """
    Run objdump on a single binary and extract the range for a given symbol.

    Output is printed to stdout.

    Args:
        binary (str): The path of the binary file to process.
        symbol (str): The symbol to match.

    Raises:
        CalledProcessException: The ``objdump`` subprocess failed for any reason.
    """
    args = [
        "objdump", "-C",
        "-M", "intel",
        "--no-show-raw",
        "-d", "-S",
        binary
    ]

    result = sp.run(args, stdout=sp.PIPE, stderr=sp.PIPE,
                    check=True, universal_newlines=True)

    funcPattern = re.compile(r"^[0-9a-f]{16} <(.*?)\(.*\)>:$")

    funcLines = []
    funcActive = False
    lines = result.stdout.splitlines()

    for line in lines:
        match = funcPattern.match(line)
        if match:
            funcName = match.group(1)
            if funcName == symbol:
                funcActive = True
            else:
                funcActive = False

        if funcActive:
            funcLines.append(line)

    print("\n".join(funcLines))

def parse_command_line():
    """
    Parse the command line.

    Returns:
        Namespace: The parsed command line container.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("binary", type=argparse.FileType("r"),
                        help="The new binary to dump")

    parser.add_argument("symbol", type=str,
                        help="The function name to dump")

    return parser.parse_args()


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    args = parse_command_line()

    # Preflight - check that size exists. Note that size might still fail at
    # runtime later, e.g. if the binary is not of the correct format
    path = shutil.which("objdump")
    if not path:
        print("ERROR: The 'objdump' utility is not installed on the PATH")
        return 1

    # Collect the data
    try:
        run_objdump(args.binary.name, args.symbol)
    except sp.CalledProcessError as ex:
        print("ERROR: The 'objdump' utility failed")
        print("       %s" % ex.stderr.strip())
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2021-2022 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The ``astc_quality_test`` utility provides a tool to sweep quality settings.
"""

import numpy as np
import re
import subprocess as sp
import sys


def get_psnr_pattern():
    return r"\s*PSNR \(LDR-RGB\):\s*([0-9.]*) dB"


def get_coding_rate_pattern():
    return r"\s*Coding rate:\s*([0-9.]*) MT/s"


def parse_output(output):
    # Regex pattern for image quality
    patternPSNR = re.compile(get_psnr_pattern())
    patternCRate = re.compile(get_coding_rate_pattern())

    # Extract results from the log
    runPSNR = None
    runCRate = None

    for line in output:
        match = patternPSNR.match(line)
        if match:
            runPSNR = float(match.group(1))

        match = patternCRate.match(line)
        if match:
            runCRate = float(match.group(1))

    assert runPSNR is not None, "No coding PSNR found"
    assert runCRate is not None, "No coding rate found"
    return (runPSNR, runCRate)

def execute(command):
    """
    Run a subprocess with the specified command.

    Args:
        command (list(str)): The list of command line arguments.

    Returns:
        list(str): The output log (stdout) split into lines.
    """
    try:
        result = sp.run(command, stdout=sp.PIPE, stderr=sp.PIPE,
                        check=True, universal_newlines=True)
    except (OSError, sp.CalledProcessError):
        print("ERROR: Test run failed")
        print("  + %s" % " ".join(command))
        qcommand = ["\"%s\"" % x for x in command]
        print("  + %s" % ", ".join(qcommand))
        sys.exit(1)

    return result.stdout.splitlines()

def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    for block in ("4x4", "5x5", "6x6", "8x8", "10x10"):

        for quality in range (0, 101, 2):

            resultsQ = []
            resultsS = []

            if (quality < 40):
                repeats = 20
            elif (quality < 75):
                repeats = 10
            else:
                repeats = 5

            for _ in range(0, repeats):
                command = [
                    "./bin/astcenc-avx2",
                    "-tl",
                    "./Test/Images/Kodak/LDR-RGB/ldr-rgb-kodak23.png",
                    "/dev/null",
                    block,
                    "%s" % quality,
                    "-silent"
                ]

                stdout = execute(command)
                psnr, mts = parse_output(stdout)
                resultsQ.append(psnr)
                resultsS.append(mts)

            print("%s, %u, %0.3f, %0.3f" % (block, quality, np.mean(resultsS), np.mean(resultsQ)))


    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2020-2021 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The ``astc_test_result_report.py`` script consolidates all current sets of
reference results into a single report giving PSNR diffs (absolute) and
performance diffs (relative speedup, 1 = no change).
"""

import re
import os
import sys


import testlib.resultset as trs
from collections import defaultdict as ddict


CONFIG_FILTER = [
    re.compile(r"^.*1\.7.*$"),
    re.compile(r"^.*sse.*$")
]

TESTSET_FILTER = [
    re.compile(r"^Small$"),
    re.compile(r"^Frymire$"),
]

QUALITY_FILTER = [
]

BLOCKSIZE_FILTER = [
    re.compile(r"^12x12$")
]


def find_reference_results():
    """
    Scrape the Test/Images directory for result CSV files and return an
    mapping of the result sets.

    Returns:
        Returns a three deep tree of dictionaries, with the final dict
        pointing at a `ResultSet` object. The hierarchy is:

            imageSet => quality => encoder => result
    """
    scriptDir = os.path.dirname(__file__)
    imageDir = os.path.join(scriptDir, "Images")

    # Pattern for extracting useful data from the CSV file name
    filePat = re.compile(r"astc_reference-(.+)_(.+)_results\.csv")

    # Build a three level dictionary we can write into
    results = ddict(lambda: ddict(lambda: ddict()))

    # Final all CSVs, load them and store them in the dict tree
    for root, dirs, files in os.walk(imageDir):
        for name in files:
            match = filePat.match(name)
            if match:

                # Skip results set in the filter
                skip = [1 for filt in CONFIG_FILTER if filt.match(name)]
                if skip:
                    continue

                fullPath = os.path.join(root, name)

                encoder = match.group(1)
                quality = match.group(2)
                imageSet = os.path.basename(root)

                # Skip results set in the filter
                skip = [1 for filt in TESTSET_FILTER if filt.match(imageSet)]
                if skip:
                    continue

                # Skip results set in the filter
                skip = [1 for filt in QUALITY_FILTER if filt.match(quality)]
                if skip:
                    continue

                testRef = trs.ResultSet(imageSet)
                testRef.load_from_file(fullPath)

                patchedRef = trs.ResultSet(imageSet)

                for result in testRef.records:
                    skip = [1 for filt in BLOCKSIZE_FILTER if filt.match(result.blkSz)]
                    if not skip:
                        patchedRef.add_record(result)

                results[imageSet][quality]["ref-%s" % encoder] = patchedRef

    return results


class DeltaRecord():
    """
    Record a single image result from N different encoders.

    Attributes:
        imageSet: The image set this cme from.
        quality: The compressor quality used.
        encoders: The names of the encoders used. The first encoder in this
            list will be used as the reference result.
        records: The raw records for the encoders. The order of records in this
            list matches the order of the `encoders` list.
    """

    def __init__(self, imageSet, quality, encoders, records):
        self.imageSet = imageSet
        self.quality = quality

        self.encoders = list(encoders)
        self.records = list(records)

        assert(len(self.encoders) == len(self.records))

    def get_delta_header(self, tag):
        """
        Get the delta encoding header.

        Args:
            tag: The field name to include in the tag.

        Return:
            The array of strings, providing the header names.
        """
        result = []

        for encoder in self.encoders[1:]:
            result.append("%s %s" % (tag, encoder))

        return result

    def get_abs_delta(self, field):
        """
        Get an absolute delta result.

        Args:
            field: The Record attribute name to diff.

        Return:
            The array of float delta values.
        """
        result = []

        root = self.records[0]
        for record in self.records[1:]:
            result.append(getattr(record, field) - getattr(root, field))

        return result

    def get_rel_delta(self, field):
        """
        Get an relative delta result (score / ref).

        Args:
            field: The Record attribute name to diff.

        Return:
            The array of float delta values.
        """
        result = []

        root = self.records[0]
        for record in self.records[1:]:
            result.append(getattr(record, field) / getattr(root, field))

        return result

    def get_irel_delta(self, field):
        """
        Get an inverse relative delta result (ref / score).

        Args:
            field: The Record attribute name to diff.

        Return:
            The array of float delta values.
        """
        return [1.0 / x for x in self.get_rel_delta(field)]

    def get_full_row_header_csv(self):
        """
        Get a CSV encoded delta record header.

        Return:
            The string for the row.
        """
        rows = [
            "Image Set",
            "Quality",
            "Size",
            "Name"
        ]

        rows.append("")
        rows.extend(self.get_delta_header("PSNR"))

        rows.append("")
        rows.extend(self.get_delta_header("Speed"))

        return ",".join(rows)

    def get_full_row_csv(self):
        """
        Get a CSV encoded delta record.

        Return:
            The string for the row.
        """
        rows = [
            self.imageSet,
            self.quality,
            self.records[0].name,
            self.records[0].blkSz
        ]

        rows.append("")
        data = ["%0.3f" % x for x in self.get_abs_delta("psnr")]
        rows.extend(data)

        rows.append("")
        data = ["%0.3f" % x for x in self.get_irel_delta("cTime")]
        rows.extend(data)

        return ",".join(rows)


def print_result_set(imageSet, quality, encoders, results, printHeader):
    """
    Attributes:
        imageSet: The image set name.
        quality: The compressor quality used.
        encoders: The names of the encoders used. The first encoder in this
            list will be used as the reference result.
        results: The dict of results, indexed by encoder.
        printHeader: True if the table header should be printed, else False.
    """
    results = [results[x] for x in encoders]
    recordSizes = [len(x.records) for x in results]

    # Skip result sets that are not the same size
    # TODO: We can take the set intersection here to report what we can
    if min(recordSizes) != max(recordSizes):
        return

    # Interleave all result records
    recordSets = zip(*[x.records for x in results])

    # Iterate each image
    for recordSet in recordSets:
        base = recordSet[0]

        # Sanity check consistency
        for record in recordSet[1:]:
            assert(record.blkSz == base.blkSz)
            assert(record.name == base.name)

        dr = DeltaRecord(imageSet, quality, encoders, recordSet)

        if printHeader:
            print(dr.get_full_row_header_csv())
            printHeader = False

        print(dr.get_full_row_csv())


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """

    results = find_reference_results()

    imageSet = sorted(results.keys())

    first = True
    for image in imageSet:
        qualityTree = results[image]
        qualitySet = sorted(qualityTree.keys())

        for qual in qualitySet:
            encoderTree = qualityTree[qual]
            encoderSet = sorted(encoderTree.keys())

            if len(encoderSet) > 1:
                print_result_set(image, qual, encoderSet, encoderTree, first)
                first = False

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2020-2022 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The ``astc_test_result_plot.py`` script consolidates all current sets of
reference results into a single graphical plot.
"""

import re
import os
import sys

import numpy as np
import matplotlib.pyplot as plt

import testlib.resultset as trs
from collections import defaultdict as ddict


def find_reference_results():
    """
    Scrape the Test/Images directory for result CSV files and return an
    mapping of the result sets.

    Returns:
        Returns a three deep tree of dictionaries, with the final dict
        pointing at a `ResultSet` object. The hierarchy is:

            imageSet => quality => encoder => result
    """
    scriptDir = os.path.dirname(__file__)
    imageDir = os.path.join(scriptDir, "Images")

    # Pattern for extracting useful data from the CSV file name
    filePat = re.compile(r"astc_reference-(.+)_(.+)_results\.csv")

    # Build a three level dictionary we can write into
    results = ddict(lambda: ddict(lambda: ddict()))

    # Final all CSVs, load them and store them in the dict tree
    for root, dirs, files in os.walk(imageDir):
        for name in files:
            match = filePat.match(name)
            if match:
                fullPath = os.path.join(root, name)

                encoder = match.group(1)
                quality = match.group(2)
                imageSet = os.path.basename(root)

                if imageSet not in ["Kodak", "Khronos", "HDRIHaven", "KodakSim"]:
                    continue

                testRef = trs.ResultSet(imageSet)
                testRef.load_from_file(fullPath)

                results[imageSet][quality]["ref-%s" % encoder] = testRef

    return results


def get_series(results, tgtEncoder, tgtQuality, resFilter=lambda x: True):
    psnrData = []
    mtsData = []
    marker = []
    records = []

    for imageSet, iResults in results.items():

        for quality, qResults in iResults.items():
            if quality != tgtQuality:
                continue

            for encoder, eResults in qResults.items():
                if encoder != tgtEncoder:
                    continue

                for record in eResults.records:
                    if resFilter(record):
                        records.append(record)
                        psnrData.append(record.psnr)
                        mtsData.append(record.cRate)

                        if "ldr-xy" in record.name:
                            marker.append('$N$')
                        elif "ldr-l" in record.name:
                            marker.append('$G$')
                        elif "ldr" in record.name:
                            marker.append('$L$')
                        elif "hdr" in record.name:
                            marker.append('$H$')
                        else:
                            marker.append('$?$')


    return mtsData, psnrData, marker, records


def get_series_rel(results, refEncoder, refQuality, tgtEncoder, tgtQuality, resFilter=lambda x: True):

    mts1, psnr1, marker1, rec1 = get_series(results, tgtEncoder, tgtQuality, resFilter)

    if refEncoder is None:
        refEncoder = tgtEncoder

    if refQuality is None:
        refQuality = tgtQuality

    mts2, psnr2, marker2, rec2 = get_series(results, refEncoder, refQuality, resFilter)

    mtsm  = [x/mts2[i] for i, x in enumerate(mts1)]
    psnrm = [x - psnr2[i] for i, x in enumerate(psnr1)]

    return mtsm, psnrm, marker1, rec1


def get_human_eq_name(encoder, quality):
    parts = encoder.split("-")
    if len(parts) == 2:
        return "astcenc %s -%s" % (parts[1], quality)
    else:
        return "astcenc-%s %s -%s" % (parts[2], parts[1], quality)


def get_human_e_name(encoder):
    parts = encoder.split("-")
    if len(parts) == 2:
        return "astcenc %s" % parts[1]
    else:
        return "astcenc-%s %s" % (parts[2], parts[1])


def get_human_q_name(quality):
    return "-%s" % quality


def plot(results, chartRows, chartCols, blockSizes,
         relative, pivotEncoder, pivotQuality, fileName, limits):

    fig, axs = plt.subplots(nrows=len(chartRows), ncols=len(chartCols),
                            sharex=True, sharey=True, figsize=(15, 8.43))

    for a in fig.axes:
        a.tick_params(
            axis="x", which="both",
            bottom=True, top=False, labelbottom=True)

        a.tick_params(
            axis="y", which="both",
            left=True, right=False, labelleft=True)

    for i, row in enumerate(chartRows):
        for j, col in enumerate(chartCols):
            if row == "fastest" and (("1.7" in col) or ("2.0" in col)):
                if len(chartCols) == 1:
                    fig.delaxes(axs[i])
                else:
                    fig.delaxes(axs[i][j])
                continue

            if len(chartRows) == 1 and len(chartCols) == 1:
                ax = axs
            elif len(chartCols) == 1:
                ax = axs[i]
            else:
                ax = axs[i, j]

            title = get_human_eq_name(col, row)

            if not relative:
                ax.set_title(title, y=0.97, backgroundcolor="white")
                ax.set_xlabel('Coding performance (MTex/s)')
                ax.set_ylabel('PSNR (dB)')
            else:
                if pivotEncoder and pivotQuality:
                    tag = get_human_eq_name(pivotEncoder, pivotQuality)
                elif pivotEncoder:
                    tag = get_human_e_name(pivotEncoder)
                else:
                    assert(pivotQuality)
                    tag = get_human_q_name(pivotQuality)

                ax.set_title("%s vs. %s" % (title, tag), y=0.97, backgroundcolor="white")
                ax.set_xlabel('Performance scaling')
                ax.set_ylabel('PSNR delta (dB)')

            for k, series in enumerate(blockSizes):
                fn = lambda x: x.blkSz == series

                if not relative:
                    x, y, m, r = get_series(results, col, row, fn)
                else:
                    x, y, m, r = get_series_rel(results, pivotEncoder, pivotQuality,
                                                col, row, fn)

                color = None
                label = "%s blocks" % series
                for xp, yp, mp in zip(x, y, m):
                    ax.scatter([xp],[yp], s=16, marker=mp,
                               color="C%u" % k, label=label)
                    label = None

            if i == 0 and j == 0:
                ax.legend(loc="lower right")

    for i, row in enumerate(chartRows):
        for j, col in enumerate(chartCols):

            if len(chartRows) == 1 and len(chartCols) == 1:
                ax = axs
            elif len(chartCols) == 1:
                ax = axs[i]
            else:
                ax = axs[i, j]

            ax.grid(ls=':')

            if limits and limits[0]:
                ax.set_xlim(left=limits[0][0], right=limits[0][1])
            if limits and limits[1]:
                ax.set_ylim(bottom=limits[1][0], top=limits[1][1])

    fig.tight_layout()
    fig.savefig(fileName)


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    absXMin = 0
    absXMax = 80
    absXLimits = (absXMin, absXMax)

    relXMin = 0.8
    relXMax = None
    relXLimits = (relXMin, relXMax)

    last1x = "1.7"
    last2x = "2.5"
    last3x = "3.7"
    prev4x = "4.3"
    last4x = "4.4"
    lastMain = "main"

    charts = [
        # --------------------------------------------------------
        # Latest in stable series charts
        [
            # Relative scores
            ["thorough", "medium", "fast"],
            [f"ref-{last2x}-avx2", f"ref-{last3x}-avx2", f"ref-{last4x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last1x}",
            None,
            "results-relative-stable-series.png",
            (None, None)
        ], [
            # Absolute scores
            ["thorough", "medium", "fast"],
            [f"ref-{last1x}", f"ref-{last2x}-avx2", f"ref-{last3x}-avx2", f"ref-{last4x}-avx2"],
            ["4x4", "6x6", "8x8"],
            False,
            None,
            None,
            "results-absolute-stable-series.png",
            (absXLimits, None)
        ],
        # --------------------------------------------------------
        # Latest 2.x vs 1.x release charts
        [
            # Relative scores
            ["thorough", "medium", "fast"],
            [f"ref-{last2x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last1x}",
            None,
            "results-relative-2.x-vs-1.x.png",
            (None, None)
        ],
        # --------------------------------------------------------
        # Latest 3.x vs 1.x release charts
        [
            # Relative scores
            ["thorough", "medium", "fast"],
            [f"ref-{last3x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last1x}",
            None,
            "results-relative-3.x-vs-1.x.png",
            (None, None)
        ],
        # --------------------------------------------------------
        # Latest 4.x vs 1.x release charts
        [
            # Relative scores
            ["thorough", "medium", "fast"],
            [f"ref-{last4x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last1x}",
            None,
            "results-relative-4.x-vs-1.x.png",
            (None, None)
        ],
        # --------------------------------------------------------
        # Latest 3.x vs 2.x release charts
        [
            # Relative scores
            ["thorough", "medium", "fast", "fastest"],
            [f"ref-{last3x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last2x}-avx2",
            None,
            "results-relative-3.x-vs-2.x.png",
            (None, None)
        ],
        # --------------------------------------------------------
        # Latest 4.x vs 3.x release charts
        [
            # Relative scores
            ["thorough", "medium", "fast", "fastest"],
            [f"ref-{last4x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last3x}-avx2",
            None,
            "results-relative-4.x-vs-3.x.png",
            (relXLimits, None),
        ], [
            # Relative ISAs of latest
            ["thorough", "medium", "fast", "fastest"],
            [f"ref-{last4x}-sse4.1", f"ref-{last4x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last4x}-sse2",
            None,
            "results-relative-4.x-isa.png",
            (None, None)
        ], [
            # Relative quality of latest
            ["medium", "fast", "fastest"],
            [f"ref-{last4x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            None,
            "thorough",
            "results-relative-4.x-quality.png",
            (None, None)
        ],
        # --------------------------------------------------------
        # Latest 4.x vs previous 4.x release charts
        [
            # Relative scores
            ["thorough", "medium", "fast", "fastest"],
            [f"ref-{last4x}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{prev4x}-avx2",
            None,
            "results-relative-4.x-vs-4.x.png",
            (relXLimits, None)
        ],
        # --------------------------------------------------------
        # Latest 4.x vs previous 4.x release charts
        [
            # Relative scores
            ["thorough", "medium", "fast", "fastest"],
            [f"ref-{lastMain}-avx2"],
            ["4x4", "6x6", "8x8"],
            True,
            f"ref-{last4x}-avx2",
            None,
            "results-relative-main-vs-4.x.png",
            (relXLimits, None)
        ]
    ]

    results = find_reference_results()

    # Force select is triggered by adding a trailing entry to the argument list
    # of the charts that you want rendered; designed for debugging use cases
    maxIndex = 0
    expectedLength = 8
    for chart in charts:
        maxIndex = max(maxIndex, len(chart))

    for chart in charts:
        # If force select is enabled then only keep the forced ones
        if len(chart) != maxIndex:
            print("Skipping %s" % chart[6])
            continue
        else:
            print("Generating %s" % chart[6])

        # If force select is enabled then strip the dummy force option
        if maxIndex != expectedLength:
            chart = chart[:expectedLength]

        plot(results, *chart)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2019-2024 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The image test runner is used for image quality and performance testing.

It is designed to process directories of arbitrary test images, using the
directory structure and path naming conventions to self-describe how each image
is to be compressed. Some built-in test sets are provided in the ./Test/Images
directory, and others can be downloaded by running the astc_test_image_dl
script.

Attributes:
    RESULT_THRESHOLD_WARN: The result threshold (dB) for getting a WARN.
    RESULT_THRESHOLD_FAIL: The result threshold (dB) for getting a FAIL.
    TEST_BLOCK_SIZES: The block sizes we can test. This is a subset of the
        block sizes supported by ASTC, simply to keep test run times
        manageable.
"""

import argparse
import os
import platform
import sys

import testlib.encoder as te
import testlib.testset as tts
import testlib.resultset as trs

# Require bit exact with reference scores
RESULT_THRESHOLD_WARN = -0.00
RESULT_THRESHOLD_FAIL = -0.00
RESULT_THRESHOLD_3D_FAIL = -0.00


TEST_BLOCK_SIZES = ["4x4", "5x5", "6x6", "8x8", "12x12", "3x3x3", "6x6x6"]

TEST_QUALITIES = ["fastest", "fast", "medium", "thorough", "verythorough", "exhaustive"]


def is_3d(blockSize):
    """
    Is the given block size a 3D block type?

    Args:
        blockSize (str): The block size.

    Returns:
        bool: ``True`` if the block string is a 3D block size, ``False`` if 2D.
    """
    return blockSize.count("x") == 2


def count_test_set(testSet, blockSizes):
    """
    Count the number of test executions needed for a test set.

    Args:
        testSet (TestSet): The test set to run.
        blockSizes (list(str)): The block sizes to run.

    Returns:
        int: The number of test executions needed.
    """
    count = 0
    for blkSz in blockSizes:
        for image in testSet.tests:
            # 3D block sizes require 3D images
            if is_3d(blkSz) != image.is3D:
                continue

            count += 1

    return count


def determine_result(image, reference, result):
    """
    Determine a test result against a reference and thresholds.

    Args:
        image (TestImage): The image being compressed.
        reference (Record): The reference result to compare against.
        result (Record): The test result.

    Returns:
        Result: The result code.
    """
    dPSNR = result.psnr - reference.psnr

    if (dPSNR < RESULT_THRESHOLD_FAIL) and (not image.is3D):
        return trs.Result.FAIL

    if (dPSNR < RESULT_THRESHOLD_3D_FAIL) and image.is3D:
        return trs.Result.FAIL

    if dPSNR < RESULT_THRESHOLD_WARN:
        return trs.Result.WARN

    return trs.Result.PASS


def format_solo_result(image, result):
    """
    Format a metrics string for a single (no compare) result.

    Args:
        image (TestImage): The image being tested.
        result (Record): The test result.

    Returns:
        str: The metrics string.
    """
    name = "%5s %s" % (result.blkSz, result.name)
    tPSNR = "%2.3f dB" % result.psnr
    tTTime = "%.3f s" % result.tTime
    tCTime = "%.3f s" % result.cTime
    tCMTS = "%.3f MT/s" % result.cRate

    return "%-32s | %8s | %9s | %9s | %11s" % \
        (name, tPSNR, tTTime, tCTime, tCMTS)


def format_result(image, reference, result):
    """
    Format a metrics string for a comparison result.

    Args:
        image (TestImage): The image being tested.
        reference (Record): The reference result to compare against.
        result (Record): The test result.

    Returns:
        str: The metrics string.
    """
    dPSNR = result.psnr - reference.psnr

    try:
        sTTime = reference.tTime / result.tTime
    except ZeroDivisionError:
        sTTime = float('NaN')

    try:
        sCTime = reference.cTime / result.cTime
    except ZeroDivisionError:
        sCTime = float('NaN')

    name  = "%5s %s" % (result.blkSz, result.name)
    tPSNR = "%2.3f dB (% 1.3f dB)" % (result.psnr, dPSNR)
    tTTime = "%.3f s (%1.2fx)" % (result.tTime, sTTime)
    tCTime = "%.3f s (%1.2fx)" % (result.cTime, sCTime)
    tCMTS = "%.3f MT/s" % (result.cRate)
    result = determine_result(image, reference, result)

    return "%-32s | %22s | %15s | %15s | %11s | %s" % \
           (name, tPSNR, tTTime, tCTime, tCMTS, result.name)


def run_test_set(encoder, testRef, testSet, quality, blockSizes, testRuns,
                 keepOutput, threads):
    """
    Execute all tests in the test set.

    Args:
        encoder (EncoderBase): The encoder to use.
        testRef (ResultSet): The test reference results.
        testSet (TestSet): The test set.
        quality (str): The quality level to execute the test against.
        blockSizes (list(str)): The block sizes to execute each test against.
        testRuns (int): The number of test repeats to run for each image test.
        keepOutput (bool): Should the test preserve output images? This is
            only a hint and discarding output may be ignored if the encoder
            version used can't do it natively.
        threads (int or None): The thread count to use.

    Returns:
        ResultSet: The test results.
    """
    resultSet = trs.ResultSet(testSet.name)

    curCount = 0
    maxCount = count_test_set(testSet, blockSizes)

    dat = (testSet.name, encoder.name, quality)
    title = "Test Set: %s / Encoder: %s -%s" % dat
    print(title)
    print("=" * len(title))

    for blkSz in blockSizes:
        for image in testSet.tests:
            # 3D block sizes require 3D images
            if is_3d(blkSz) != image.is3D:
                continue

            curCount += 1

            dat = (curCount, maxCount, blkSz, image.testFile)
            print("Running %u/%u %s %s ... " % dat, end='', flush=True)
            res = encoder.run_test(image, blkSz, "-%s" % quality, testRuns,
                                   keepOutput, threads)
            res = trs.Record(blkSz, image.testFile, res[0], res[1], res[2], res[3])
            resultSet.add_record(res)

            if testRef:
                refResult = testRef.get_matching_record(res)
                res.set_status(determine_result(image, refResult, res))

                try:
                    res.tTimeRel = refResult.tTime / res.tTime
                except ZeroDivisionError:
                    res.tTimeRel = float('NaN')

                try:
                    res.cTimeRel = refResult.cTime / res.cTime
                except ZeroDivisionError:
                    res.cTimeRel = float('NaN')

                res.psnrRel = res.psnr - refResult.psnr

                res = format_result(image, refResult, res)
            else:
                res = format_solo_result(image, res)

            print("\r[%3u] %s" % (curCount, res))

    return resultSet


def get_encoder_params(encoderName, referenceName, imageSet):
    """
    The the encoder and image set parameters for a test run.

    Args:
        encoderName (str): The encoder name.
        referenceName (str): The reference encoder name.
        imageSet (str): The test image set.

    Returns:
        tuple(EncoderBase, str, str, str): The test parameters for the
        requested encoder and test set. An instance of the encoder wrapper
        class, the output data name, the output result directory, and the
        reference to use.
    """
    # 1.7 variants
    if encoderName == "ref-1.7":
        encoder = te.Encoder1_7()
        name = "reference-1.7"
        outDir = "Test/Images/%s" % imageSet
        refName = None
        return (encoder, name, outDir, refName)

    if encoderName.startswith("ref"):
        _, version, simd = encoderName.split("-")

        # 2.x, 3.x, and 4.x variants
        compatible2xPrefixes = ["2.", "3.", "4.", "5."]
        if any(True for x in compatible2xPrefixes if version.startswith(x)):
            encoder = te.Encoder2xRel(version, simd)
            name = f"reference-{version}-{simd}"
            outDir = "Test/Images/%s" % imageSet
            refName = None
            return (encoder, name, outDir, refName)

        # Latest main
        if version == "main":
            encoder = te.Encoder2x(simd)
            name = f"reference-{version}-{simd}"
            outDir = "Test/Images/%s" % imageSet
            refName = None
            return (encoder, name, outDir, refName)

        assert False, f"Encoder {encoderName} not recognized"

    encoder = te.Encoder2x(encoderName)
    name = "develop-%s" % encoderName
    outDir = "TestOutput/%s" % imageSet
    refName = referenceName.replace("ref", "reference")
    return (encoder, name, outDir, refName)


def parse_command_line():
    """
    Parse the command line.

    Returns:
        Namespace: The parsed command line container.
    """
    parser = argparse.ArgumentParser()

    # All reference encoders
    refcoders = ["ref-1.7",
                 "ref-2.5-neon", "ref-2.5-sse2", "ref-2.5-sse4.1", "ref-2.5-avx2",
                 "ref-3.7-neon", "ref-3.7-sse2", "ref-3.7-sse4.1", "ref-3.7-avx2",
                 "ref-4.8-neon", "ref-4.8-sse2", "ref-4.8-sse4.1", "ref-4.8-avx2",
                 "ref-5.0-neon", "ref-5.0-sse2", "ref-5.0-sse4.1", "ref-5.0-avx2",
                 "ref-main-neon", "ref-main-sve_256", "ref-main-sve_128", "ref-main-sse2", "ref-main-sse4.1", "ref-main-avx2"]

    # All test encoders
    testcoders = ["none", "neon", "sve_256", "sve_128", "sse2", "sse4.1", "avx2", "native", "universal"]
    testcodersAArch64 = ["neon", "sve_256", "sve_128"]
    testcodersX86 = ["sse2", "sse4.1", "avx2"]

    coders = refcoders + testcoders + ["all-aarch64", "all-x86"]

    parser.add_argument("--encoder", dest="encoders", default="avx2",
                        choices=coders, help="test encoder variant")

    parser.add_argument("--reference", dest="reference", default="ref-main-avx2",
                        choices=refcoders, help="reference encoder variant")

    astcProfile = ["ldr", "ldrs", "hdr", "all"]
    parser.add_argument("--color-profile", dest="profiles", default="all",
                        choices=astcProfile, help="test color profile")

    imgFormat = ["l", "xy", "rgb", "rgba", "all"]
    parser.add_argument("--color-format", dest="formats", default="all",
                        choices=imgFormat, help="test color format")

    choices = list(TEST_BLOCK_SIZES) + ["all"]
    parser.add_argument("--block-size", dest="blockSizes",
                        action="append", choices=choices,
                        help="test block size")

    testDir = os.path.dirname(__file__)
    testDir = os.path.join(testDir, "Images")
    testSets = []
    for path in os.listdir(testDir):
        fqPath = os.path.join(testDir, path)
        if os.path.isdir(fqPath):
            testSets.append(path)
    testSets.append("all")

    parser.add_argument("--test-set", dest="testSets", default="Small",
                        choices=testSets, help="test image test set")

    parser.add_argument("--test-image", dest="testImage", default=None,
                        help="select a specific test image from the test set")

    choices = list(TEST_QUALITIES) + ["all", "all+"]
    parser.add_argument("--test-quality", dest="testQual", default="thorough",
                        choices=choices, help="select a specific test quality")

    parser.add_argument("--repeats", dest="testRepeats", default=1,
                        type=int, help="test iteration count")

    parser.add_argument("--keep-output", dest="keepOutput", default=False,
                        action="store_true", help="keep image output")

    parser.add_argument("-j", dest="threads", default=None,
                        type=int, help="thread count")


    args = parser.parse_args()

    # Turn things into canonical format lists
    if args.encoders == "all-aarch64":
        args.encoders = testcodersAArch64
    elif args.encoders == "all-x86":
        args.encoders = testcodersX86
    else:
        args.encoders = [args.encoders]

    if args.testQual == "all+":
        args.testQual = TEST_QUALITIES
    elif args.testQual == "all":
        args.testQual = TEST_QUALITIES
        args.testQual.remove("verythorough")
        args.testQual.remove("exhaustive")
    else:
        args.testQual = [args.testQual]

    if not args.blockSizes or ("all" in args.blockSizes):
        args.blockSizes = TEST_BLOCK_SIZES

    args.testSets = testSets[:-1] if args.testSets == "all" \
        else [args.testSets]

    args.profiles = astcProfile[:-1] if args.profiles == "all" \
        else [args.profiles]

    args.formats = imgFormat[:-1] if args.formats == "all" \
        else [args.formats]

    return args


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    # Parse command lines
    args = parse_command_line()

    testSetCount = 0
    worstResult = trs.Result.NOTRUN

    for quality in args.testQual:
        for imageSet in args.testSets:
            for encoderName in args.encoders:
                (encoder, name, outDir, refName) = \
                    get_encoder_params(encoderName, args.reference, imageSet)

                testDir = "Test/Images/%s" % imageSet
                testRes = "%s/astc_%s_%s_results.csv" % (outDir, name, quality)

                testRef = None
                if refName:
                    dat = (testDir, refName, quality)
                    testRefPath = "%s/astc_%s_%s_results.csv" % dat
                    testRef = trs.ResultSet(imageSet)
                    testRef.load_from_file(testRefPath)

                testSetCount += 1
                testSet = tts.TestSet(imageSet, testDir,
                                      args.profiles, args.formats, args.testImage)

                resultSet = run_test_set(encoder, testRef, testSet, quality,
                                         args.blockSizes, args.testRepeats,
                                         args.keepOutput, args.threads)

                resultSet.save_to_file(testRes)

                if refName:
                    summary = resultSet.get_results_summary()
                    worstResult = max(summary.get_worst_result(), worstResult)
                    print(summary)

        if (testSetCount > 1) and (worstResult != trs.Result.NOTRUN):
            print("OVERALL STATUS: %s" % worstResult.name)

    if worstResult == trs.Result.FAIL:
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
# Copyright 2020 Arm Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -----------------------------------------------------------------------------
"""
The python test runner is designed to run some basic tests against the Python
test code base.
"""

import re
import sys
import unittest

import pycodestyle
import pylint.epylint as lint


class PythonTests(unittest.TestCase):
    """
    Some basic Python static analysis and style checks.
    """

    def test_pylint(self):
        """
        Run pylint over the codebase.
        """
        pylintOut, _ = lint.py_run("./Test", True)
        pattern = re.compile(r"Your code has been rated at (.*?)/10")
        match = pattern.search(pylintOut.getvalue())
        self.assertIsNotNone(match)
        score = float(match.group(1))
        self.assertGreaterEqual(score, 9.8)

        with open("pylint.log", "w") as fileHandle:
            fileHandle.write(pylintOut.getvalue())

    def test_pycodestyle(self):
        """
        Test that we conform to PEP-8.
        """
        style = pycodestyle.StyleGuide()
        result = style.check_files(["./Test"])
        self.assertEqual(result.total_errors, 0,
                         "Found code style errors (and warnings).")


def main():
    """
    The main function.

    Returns:
        int: The process return code.
    """
    results = unittest.main(exit=False)
    return 0 if results.result.wasSuccessful() else 1


if __name__ == "__main__":
    sys.exit(main())
